/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#ifndef WRONCEFS_H
#define WRONCEFS_H

#include "wroncefs_config.h"
#include <stdint.h>

/* this macro is placed before all function forward declarations, can be */
/* overridden in "wroncefs_config.h" */
#ifndef WOFS_EXTERN
#  define WOFS_EXTERN extern
#endif /* WOFS_EXTERN */

/* this macro is placed before all function definitions, can be overridden 
/* in "wroncefs_config.h" */
#ifndef WOFS_IMPL
#  define WOFS_IMPL
#endif /* WOFS_IMPL */

#if WRONCEFS_USER_STATE
#  define WOFS_STATE_ARG struct wofs_state* const state,
#else /* WRONCEFS_USER_STATE */
#  define WOFS_STATE_ARG
#endif /* WRONCEFS_USER_STATE */

#define WOFS_INVALID_INODE  0
#define WOFS_ROOT_INODE     1
#define WOFS_END_INODE      0xFFFF

typedef uint16_t wofs_ino_t;

#define WOFS_MIN_BLOCKSZSH 6

#if WRONCEFS_16BIT_MODE
typedef int16_t  wofs_off_t;
typedef uint16_t wofs_addr_t;
typedef uint16_t wofs_size_t;
typedef int16_t  wofs_ssize_t;
#  define WOFS_MAX_FILESIZE  0x7FFF
#  define WOFS_MAX_BLOCKSZSH 15
#else /* WRONCEFS_16BIT_MODE */
typedef int32_t  wofs_off_t;
typedef uint32_t wofs_addr_t;
typedef uint32_t wofs_size_t;
typedef int32_t  wofs_ssize_t;
#  define WOFS_MAX_FILESIZE  0x7FFFFFFF
#  define WOFS_MAX_BLOCKSZSH 16
#endif /* WRONCEFS_16BIT_MODE */

/* we require at least 8 blocks, otherwise the filesystem does not have */
/* much sense */
#define WOFS_MIN_NBLOCKS 8

/* maximal supported block count */
#define WOFS_MAX_NBLOCKS 0xFFFE

/* maximal supported value for WRONCEFS_NAME_MAX */
#define WOFS_MAX_NAME_MAX 254

#if WRONCEFS_NAME_MAX>WOFS_MAX_NAME_MAX
#  error "WRONCEFS_NAME_MAX is too large"
#endif /* WRONCEFS_NAME_MAX>WOFS_MAX_NAME_MAX */

struct wofs_inode
{
	uint16_t   i_sig;
	wofs_ino_t i_first;
	wofs_ino_t i_next;
	uint8_t    i_vnext;
	uint8_t    i_resd;
	uint8_t    i_type;
	uint8_t    i_alive;
};

#define WOFS_INOTYPE_DEAD     0x00
#define WOFS_INOTYPE_FILE     0x1E
#define WOFS_INOTYPE_DIR      0xD1
#define WOFS_INOTYPE_TMPDIR   0x7D
#define WOFS_INOTYPE_TMPFILE  0x7E

struct wofs_fsinfo
{
	uint8_t  fsi_blockszsh;  /* block size shift */
	uint8_t  fsi_maxnamelen;
	uint16_t fsi_nblocks;
	uint16_t fsi_flags;
	uint16_t fsi_reserved;
};

#define WOFS_FSINFO_CANALTER    0x0001
#define WOFS_FSINFO_16BITOFFS   0x0002

struct wofs_dirent
{
	wofs_ino_t d_ino;
	uint8_t    d_namelen;
	char       d_name[WRONCEFS_NAME_MAX+1];
};

struct wofs_fd
{
	wofs_addr_t fd_addr; /* base address of the file inode */
	wofs_off_t  fd_fpos; /* current file position */
	uint16_t    fd_lpos; /* local (block-relative) position */
	uint16_t    fd_clen; /* data length in the current block */
#if !WRONCEFS_CAN_ALTER
	uint16_t    fd_plim; /* position of the length indicator marker */
#endif /* !WRONCEFS_CAN_ALTER */
};

struct wofs_state
{
	uint16_t nblocks;
	uint16_t freeblocks;
	uint8_t  blockszsh;
	uint8_t  name_max;
};

/* opaque type for dirent hint, used internally only */
struct wofs_direnthint;

struct wofs_dir
{
	wofs_addr_t             dir_addr; /* absolute position */
	wofs_addr_t             dir_end;  /* address of the processed inode */
	wofs_ino_t              dir_ino;  /* inode that is processed */
	wofs_size_t             dir_lost; /* space lost due to dead dirents */
	struct wofs_direnthint* dir_hint; /* reserved for the implementation */
	struct wofs_dirent      dir_ent;
};

struct wofs_statfs
{
	wofs_size_t f_bsize;
	wofs_size_t f_blocks;
	wofs_size_t f_bfree;
	wofs_size_t f_namelen;
};

struct wofs_fsckops
{
	void (*fix_chain)(WOFS_STATE_ARG const wofs_ino_t ino);
	void (*lost_inode)(WOFS_STATE_ARG const wofs_ino_t ino);
};

/* Do not erase all blocks on the storage, it is already erased. This is */
/* typical to SPI flash, due to the chip-erase capability. */
#define WOFS_FORMAT_DONTERASE     0x0001

/* Fix interrupted directory entry removal. This is essentional mount() to */
/* be successful. */
#define WOFS_MOUNT_FIXDIRENTREM   0x0001

/* Restore pending temporary file inodes. This is essentional mount() to */
/* be successful. */
#define WOFS_MOUNT_RESTRTEMPFILE  0x0002

/* Erase garbage inodes. Should be specified to get accurate free inode */
/* (or block) count after mount(). */
#define WOFS_MOUNT_ERASEGARBAGE   0x0004

#define WOFS_MOUNT_ESSENTIONAL    0x0003 /* only essentionals */
#define WOFS_MOUNT_NORMAL         0x0007 /* essentionals and recommended */

#define WOFS_SEEK_SET             0
#define WOFS_SEEK_CUR             1
#define WOFS_SEEK_END             2

#define WOFS_ERR_ENOSYS           1
#define WOFS_ERR_EBADSIG          2
#define WOFS_ERR_EBADINFO         3
#define WOFS_ERR_EINVAL           4
#define WOFS_ERR_EIOR             5
#define WOFS_ERR_EIOW             6
#define WOFS_ERR_EIOE             7
#define WOFS_ERR_ENOENT           8
#define WOFS_ERR_EEXIST           9
#define WOFS_ERR_ENOTDIR          10
#define WOFS_ERR_EFBIG            11
#define WOFS_ERR_EISDIR           12
#define WOFS_ERR_ENOTEMPTY        13
#define WOFS_ERR_ENOTRECOVERABLE  14
#define WOFS_ERR_ENOSPC           15
#define WOFS_ERR_ENAMETOOLONG     16
#define WOFS_ERR_EBADINODE        17
#define WOFS_ERR_ENOMORE          18
#define WOFS_ERR_EOVERFLOW        19
#define WOFS_ERR_ENODEV           20
#define WOFS_ERR_EUCLEAN          21

/* Converts a WOFS_ERR_... constant to string. Most of the WOFS API */
/* functions returns error code, however, the returned values are negative */
/* numbers. Therefore, before calling this function, the returned error */
/* code must be negated, i.e.: str = wofs_strerror(-wofs_format(...)) */
WOFS_EXTERN const char* wofs_strerror(int8_t err);

/* Get WOFS version number as a string in "xx.yy.zz" format, where "xx" is */
/* the major version number, "yy" is the minor version, and "zz" is the */
/* patch level. */
WOFS_EXTERN const char* wofs_get_version(void);

/* Compute the shift value of the specified size. Input value must be */
/* power-of-2. Returns an integer that satisfies: */
/*   off == 1 << return_value  */
/* Returns negative if 'off' is not power-of-2 or non-positive. */
WOFS_EXTERN int8_t wofs_off_to_shift(wofs_off_t off);

/* Create a new filesystem. Block size shift must not be less than */
/* WOFS_MIN_BLOCKSZSH and greater than WOFS_MAX_BLOCKSZSH. If this function */
/* succeeds, the filesystem is mounted and ready to use. 'flags' takes zero */
/* or more ORd WOFS_FORMAT_... values. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_format(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, uint16_t flags);

/* Mount a previously formatted and used filesystem. Arguments have the */
/* same restrictions as wofs_format(). Use 'flags' to specify what */
/* operations filesystem check should perform (see WOFS_MOUNT_...). These */
/* checks are considered quick, and most of them are essentional in order */
/* mount to succeed. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_mount(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, uint16_t flags);

/* Perform in-depth filesystem check. If this operation succeed, the */
/* filesystem is mounted, and can be used. Dimension arguments have the */
/* same restrictions as wofs_format(). The callback operation table ('ops') */
/* is optional. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_fsck(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, const struct wofs_fsckops* const ops);

/* Change the working directory. '*wd' must be a directory -- this */
/* directory considered the "current working directory". The function */
/* find the directory entry with the specified name in this working */
/* directory. If the entry is found and it is a directory, then '*wd' */
/* will be overwritten, and zero is returned; otherwise '*wd' will be */
/* intact and -WOFS_ERR_... is returned. */
WOFS_EXTERN int8_t wofs_chdir(WOFS_STATE_ARG wofs_ino_t* const wd,
	const char* const name);

/* Create a new directory in the specified directory by 'dir' with the */
/* specified name. The inode of the newly allocated is optionally returned */
/* in 'opt_ino'. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_mkdir(WOFS_STATE_ARG wofs_ino_t dir,
	const char* const name, wofs_ino_t* const opt_ino);

/* Remove a directory for the specified directory inode with the */
/* specified name. */
/* The directory must be empty (contains only "." and ".."). */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_rmdir(WOFS_STATE_ARG wofs_ino_t ino,
	const char* const name);

/* Create a directory iterator to the specified inode. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_opendir(WOFS_STATE_ARG struct wofs_dir* const dir,
	wofs_ino_t ino);

/* Read the next directory entry. Returns the next entry. Error code is */
/* returned in 'out_ec' if it is not NULL. */
WOFS_EXTERN struct wofs_dirent* wofs_readdir(WOFS_STATE_ARG
	struct wofs_dir* const dir, int8_t* const out_ec);

/* Create a new regular file in the specified directory inode with the */
/* specified name. The file descriptor must be supplied by the caller. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_creat(WOFS_STATE_ARG struct wofs_fd* const fd,
	wofs_ino_t dir, const char* const name);

/* Unlink a regular file from the specified directory inode with the */
/* specified name. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_unlink(WOFS_STATE_ARG wofs_ino_t dir,
	const char* const name);

/* Open an existing file with the specified name in the directory inode. */
/* The file descriptor must be supplied by the caller. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_open(WOFS_STATE_ARG struct wofs_fd* const fd,
	wofs_ino_t dir, const char* const name);

/* Open an existing file with the specified file inode. */
/* The file descriptor must be supplied by the caller. */
/* Returns zero on success, -WOFS_ERR_... otherwise. */
WOFS_EXTERN int8_t wofs_iopen(WOFS_STATE_ARG struct wofs_fd* const fd,
	const wofs_ino_t ino);

/* Write data to the specified file by 'fd'. File size is automatically */
/* increased on demand. File position will be updated. Short write is */
/* possible (on error or by reaching the maximal file size). Returns the */
/* number of successfully written bytes; negative on error (error code */
/* is optionally returned in 'out_ec'). */
WOFS_EXTERN wofs_ssize_t wofs_write(WOFS_STATE_ARG struct wofs_fd* const fd,
	const void* buf, wofs_size_t size, int8_t* const out_ec);

/* Read data from the file determined by the file descriptor 'fd'. File */
/* position will be updated. Short read is possible. Returns the number of */
/* successfully read bytes; negative on error (error code is optionally */
/* returned in 'out_ec'). */
WOFS_EXTERN wofs_ssize_t wofs_read(WOFS_STATE_ARG struct wofs_fd* const fd,
	void* buf, wofs_size_t size, int8_t* const out_ec);

/* Move the read/write file offset. 'whence' must be one of WOFS_SEEK_... */
/* Cannot seek beyond end-of-file (unlike POSIX lseek), and file offset */
/* will be changed in case of error (unlike POSIX lseek). Returns the */
/* resultant file offset, or negative on error (and optionally returns */
/* error code in 'out_ec'). */
WOFS_EXTERN wofs_off_t wofs_seek(WOFS_STATE_ARG struct wofs_fd* const fd,
	wofs_off_t offs, uint8_t whence, int8_t* const out_ec);

/* Allocate file space. Change the file data preservation to at least 'len' */
/* bytes. If file is already longer than 'len', no preservation will be */
/* made. This call will change file position. Returns error code. */
WOFS_EXTERN int8_t wofs_fallocate(WOFS_STATE_ARG struct wofs_fd* const fd,
	wofs_off_t len);

/* Stat inode by name. Set 'out_type' to the inode type and optionally */
/* 'out_ino' to the inode if 'name' was a valid entry in 'dir'. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs_stat(WOFS_STATE_ARG wofs_ino_t dir,
	const char* const name, uint8_t* const out_type,
	wofs_ino_t* const out_ino);

/* Stat inode. Set 'out_type' to the inode type if it was a valid inode. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs_istat(WOFS_STATE_ARG const wofs_ino_t ino,
	uint8_t* const out_type);

/* Get filesystem statistics. Returns error code. */
WOFS_EXTERN int8_t wofs_statfs(WOFS_STATE_ARG struct wofs_statfs* const st);


/***********************************************/
/* user-supplied function forward declarations */

WOFS_EXTERN int8_t wroncefs_op_write(WOFS_STATE_ARG const void* buf,
	wofs_addr_t addr, wofs_size_t size);

WOFS_EXTERN int8_t wroncefs_op_erase(WOFS_STATE_ARG wofs_addr_t addr,
	wofs_size_t size);

#if !WRONCEFS_DIRECT_READ
WOFS_EXTERN int8_t wroncefs_op_read(WOFS_STATE_ARG void* buf,
	wofs_addr_t addr, wofs_size_t size);
#endif /* !WRONCEFS_DIRECT_READ */

#endif /* WRONCEFS_H */
