/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>

WOFS_IMPL int8_t wofs__create_object(WOFS_STATE_ARG wofs_ino_t* const ino,
	wofs_ino_t dir, const char* const name, const uint8_t i_type,
	int8_t (*init_inode_callback)(WOFS_STATE_ARG wofs_addr_t addr,
	wofs_ino_t cur_ino, wofs_ino_t par_ino))
{
	int8_t ec;
	uint8_t namelen;

	if(0!=(ec=wofs__check_name(PASS_STATE name, &namelen)))
	{
		return ec;
	}

	if(0!=(ec=wofs__check_inode(PASS_STATE dir, WOFS_INOTYPE_DIR)))
	{
		return ec;
	}

	struct wofs_direnthint hint;

	{
		wofs_ino_t trg;

		ec = wofs__find_dirent(PASS_STATE &trg, dir, namelen, name, &hint);

		if(0==ec)
		{
			return -WOFS_ERR_EEXIST;
		}
		else
		if((-WOFS_ERR_ENOENT)!=ec)
		{
			return ec;
		}
	}

	struct wofs_inode inode;

	inode.i_sig   = WOFS_SIG;
	inode.i_first = 0;
	inode.i_next  = WOFS_END_INODE;
	inode.i_vnext = 0xFF;
	inode.i_resd  = 0xFF;
	inode.i_type  = i_type;

	if(0!=(ec=wofs__alloc_inode(PASS_STATE &inode.i_first)))
	{
		return ec;
	}

	const wofs_addr_t addr = ((wofs_addr_t)WOFS_INODE_TO_BLOCK(inode.i_first))
		<< WOFS_STATE.blockszsh;

	if(0!=(ec=wofs__ino_fill_1(PASS_STATE &inode, addr)))
	{
	revert_and_exit:;
		++WOFS_STATE.freeblocks;
		return ec;
	}

	if(0!=(ec=init_inode_callback(PASS_STATE addr, inode.i_first, dir)))
	{
	erase_and_fail:;
		wroncefs_op_erase(PASS_STATE addr, ((size_t)1)<<WOFS_STATE.blockszsh);
		goto revert_and_exit;
	}

	if(0!=(ec=wofs__ino_fill_2(PASS_STATE &inode, addr)))
	{
		goto erase_and_fail;
	}

	if(0!=(ec=wofs__add_dirent(PASS_STATE dir, inode.i_first, namelen, name,
		&hint)))
	{
		/* better to leave the finalized inode as (possibly) lost inode, */
		/* than randomly erase it, which could lead to corruption */
		return ec;
	}

	*ino = inode.i_first;

	return 0;
}
