/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>
#include <string.h>

WOFS_IMPL int8_t wofs__add_dirent(WOFS_STATE_ARG wofs_ino_t dir,
	const wofs_ino_t ino, const uint8_t namelen, const char* const name,
	const struct wofs_direnthint* const p_hint)
{
	/* we make a copy from hint onto the stack (or build a new one, see */
	/* below) -- hint is not large, 4-6 bytes only */
	struct wofs_direnthint hint;

	if(NULL!=p_hint)
	{
		memcpy(&hint, p_hint, sizeof(hint));
	}
	else
	{
		wofs_ino_t tmp;

		/* so, the user did not supplied a hint -- let we make one */

		const int8_t ec = wofs__find_dirent(PASS_STATE &tmp, ino, namelen,
			name, &hint);

		/* the only acceptable outcome is ENOENT, anything else (including */
		/* success!) is an error-condition */

		if((-WOFS_ERR_ENOENT)!=ec)
		{
			/* success means the entry already exists, but we about to make */
			/* a directory entry with that name */
			if(0==ec)
			{
				return -WOFS_ERR_EEXIST;
			}

			return ec;
		}
	}

	if(0==hint.addr)
	{
		/* there is no space in either existing inode -- at this point, */
		/* 'hint.ino' is the last inode in the chain, so we can append to it */

		wofs_ino_t new_ino;

		int8_t ec = wofs__alloc_inode(PASS_STATE &new_ino);

		if(0!=ec)
		{
			return ec;
		}

		const uint8_t shift = WOFS_STATE.blockszsh;

		/* set 'hint.addr' to the newly allocated space */
		hint.addr = ((wofs_addr_t)WOFS_INODE_TO_BLOCK(new_ino)) << shift;

		struct wofs_inode ino_data;

		ino_data.i_sig   = WOFS_SIG;
		ino_data.i_first = dir;
		ino_data.i_next  = WOFS_END_INODE;
		ino_data.i_type  = WOFS_INOTYPE_DIR;
		ino_data.i_vnext = 0xFF;
		ino_data.i_resd  = 0xFF;
		ino_data.i_alive = 0xFF;

		if(0!=(ec=wofs__ino_fill_1(PASS_STATE &ino_data, hint.addr)))
		{
			return ec;
		}

		if(0!=wroncefs_op_write(PASS_STATE &new_ino,
			(((wofs_addr_t)WOFS_INODE_TO_BLOCK(hint.ino)) << shift)+
			offsetof(struct wofs_inode, i_next), sizeof(new_ino)))
		{
			return -WOFS_ERR_EIOW;
		}

		if(0!=(ec=wofs__ino_fill_2(PASS_STATE &ino_data, hint.addr)))
		{
			return ec;
		}

		uint8_t vnext = 0;

		if(0!=wroncefs_op_write(PASS_STATE &vnext,
			(((wofs_addr_t)WOFS_INODE_TO_BLOCK(hint.ino)) << shift)+
			offsetof(struct wofs_inode, i_vnext), sizeof(vnext)))
		{
			return -WOFS_ERR_EIOW;
		}

		/* inode structure is complete, increment 'hint.addr' to point to */
		/* the first empty dirent slot */
		hint.addr += sizeof(ino_data);
	}

	/* we have a valid address to write the entry info */
	{
		struct wofs_rawdirent rde;

		rde.d_ino     = ino; /* the target! */
		rde.d_namelen = namelen;
		/* 'rde.d_alive' is not set, we keep it as it was (0xFF) */

		/* write dirent carefully, the write order must be: */
		/*    name -> d_ino -> d_namelen */
		/* in order to provide recovery from interrupted operations */

		if(0!=wroncefs_op_write(PASS_STATE name,
			hint.addr+sizeof(rde)-sizeof(rde.d_name2),
			namelen))
		{
			return -WOFS_ERR_EIOW;
		}

		if(0!=wroncefs_op_write(PASS_STATE &rde.d_ino,
			hint.addr+offsetof(struct wofs_rawdirent, d_ino),
			sizeof(rde.d_ino)))
		{
			return -WOFS_ERR_EIOW;
		}

		if(0!=wroncefs_op_write(PASS_STATE &rde.d_namelen,
			hint.addr+offsetof(struct wofs_rawdirent, d_namelen),
			sizeof(rde.d_namelen)))
		{
			return -WOFS_ERR_EIOW;
		}
	}

	return 0;
}
