/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"

WOFS_IMPL int8_t wofs__file_rewrite(WOFS_STATE_ARG struct wofs_fd* const fd,
	const uint16_t clen)
{
	int8_t ec;

	const uint8_t shift = WOFS_STATE.blockszsh;

	const wofs_size_t block_size = ((wofs_size_t)1) << shift;

	const wofs_ino_t orig_ino = WOFS_BLOCK_TO_INODE(fd->fd_addr >> shift);

	WOFS_READ_LOCAL(struct wofs_inode, inode, fd->fd_addr);

	struct wofs_inode temp_inode;

	if(0!=(ec=wofs__compile_temp_inode(PASS_STATE &temp_inode,
		WOFS_LOCAL_ADDR(inode), WOFS_INOTYPE_TMPFILE)))
	{
		return ec;
	}

	/* 1. Allocate a temporary file inode */

	wofs_addr_t temp_addr;

	if(0!=(ec=wofs__get_temp_block(PASS_STATE &temp_addr)))
	{
		return ec;
	}

	const wofs_ino_t temp_ino = WOFS_BLOCK_TO_INODE(temp_addr >> shift);

	/* 2. Fill the inode signature */

	if(0!=(ec=wofs__ino_fill_1(PASS_STATE &temp_inode, temp_addr)))
	{
		goto erase_and_exit;
	}

	/* 3. Copy file info (start offset, prev inode) */
	/* 4. Copy file content */

	if(0!=(ec=wofs__copy_storage_content(PASS_STATE
		temp_addr+sizeof(temp_inode), fd->fd_addr+sizeof(temp_inode),
		clen+WOFS_SIZEOF_FILEINFO)))
	{
		goto erase_and_exit;
	}

	/* 5. Write data length to the end of the block, and the original inode */
	/* right before that */

	if(0!=wroncefs_op_write(PASS_STATE &clen,
			temp_addr+block_size-sizeof(clen), sizeof(clen))
		||
		0!=wroncefs_op_write(PASS_STATE &orig_ino,
			temp_addr+block_size-sizeof(clen)-sizeof(orig_ino),
			sizeof(orig_ino)))
	{
		ec = -WOFS_ERR_EIOW;
		goto erase_and_exit;
	}

	/* 6. Finish the temporary file inode, to make it permanent */

	if(0!=(ec=wofs__ino_fill_2(PASS_STATE &temp_inode, temp_addr)))
	{
		goto erase_and_exit;
	}

	/* the rest is in the restore function */
	return wofs__file_restore_temp(PASS_STATE temp_ino);

erase_and_exit:;
	wroncefs_op_erase(PASS_STATE temp_addr, block_size);
	return ec;
}
