/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#ifndef WRONCEFS_IMPL_H
#define WRONCEFS_IMPL_H

#include "wroncefs.h"

#define WOFS_SIG 0xF557

/* we need to preserve one block in order to successfully perform the */
/* directory entry removal */
#define WOFS_NUM_SPARE_BLOCKS 1

#define WOFS_DIRENT_ALIVE   0xFF
#define WOFS_DIRENT_FSINFO  0x55
#define WOFS_DIRENT_DEAD    0x00

#define WOFS_LENGTHIND_VALID  0x18

#define WOFS_ROUND_NAMELEN(_len_)       (((_len_)+1) & (-2))
#define WOFS_INODE_TO_BLOCK(_ino_)      ((_ino_) - 1)
#define WOFS_BLOCK_TO_INODE(_block_)    ((_block_) + 1)

#define WOFS_MAX(_a_, _b_)  (((_a_)<(_b_)) ? (_b_) : (_a_))

#if WRONCEFS_USER_STATE
#  define PASS_STATE state,
#  define WOFS_STATE (*state)
#else /* WRONCEFS_USER_STATE */
#  define PASS_STATE
#  define WOFS_STATE wofs_state
#endif /* WRONCEFS_USER_STATE */

struct wofs_rawdirent
{
	wofs_ino_t d_ino;
	uint8_t    d_namelen;
	uint8_t    d_alive;
	uint8_t    d_name2[2]; /* first 2 characters of the name (lookahead) */
};

struct wofs_direnthint
{
	uint8_t     found;
	uint8_t     namelen;
	wofs_ino_t  ino;
	wofs_addr_t addr;
};

struct wofs_tempdirinfo
{
	wofs_ino_t orig_ino;  /* the original inode the data came from */
	uint16_t   cont_size; /* content size to be copied back */
};

struct wofs_fileinfo
{
	wofs_off_t start_offs; /* absolute file offset of the first data byte */
	wofs_ino_t prev_ino;   /* previous inode */
};

/* We must do this, because the compiler might round up */
/* 'struct wofs_fileinfo', and we don't want to lose space due to this. */
#define WOFS_SIZEOF_FILEINFO (sizeof(wofs_off_t)+sizeof(wofs_ino_t))

#if WRONCEFS_CAN_ALTER
struct wofs_lengthind
{
	uint16_t len1; /* the length indicator is value, if ... */
	uint16_t len2; /* ... 'len1' and 'len2' are equal */
};
#else /* WRONCEFS_CAN_ALTER */
struct wofs_lengthind
{
	uint16_t length;
	uint8_t  valid;  /* valid if hold WOFS_LENGTHIND_VALID */
};
#endif /* WRONCEFS_CAN_ALTER */

/* The minimal length of the length indicator. If WRONCEFS_CAN_ALTER=0, */
/* this size can grow, but the absolute minimal length is this value. */
#define WOFS_SIZEOF_LENGTHIND \
	(WOFS_MAX(sizeof(uint16_t), sizeof(wofs_ino_t))+sizeof(uint16_t))

WOFS_EXTERN struct wofs_state wofs_state;

WOFS_EXTERN const uint16_t wofs_fsi_flags;

/* The actual mount implementation. The separation is needed for fsck, */
/* because fsck will run between pre() and post(). Returns error code. */
WOFS_EXTERN int8_t wofs__mount_pre(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, uint16_t flags);
WOFS_EXTERN int8_t wofs__mount_post(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, uint16_t flags);

/* Generate the default directory layout (which is "." and ".."). */
/* "." will point to 'cur_ino', and ".." to 'par_ino'. Fill is performed */
/* sequentially, because its interruption will be handled by higher level */
/* mechanisms (see mkdir for instance). Returns error code. */
WOFS_EXTERN int8_t wofs__dflt_dir_layout(WOFS_STATE_ARG wofs_addr_t addr,
	wofs_ino_t cur_ino, wofs_ino_t par_ino);

/* Allocate an empty inode. All bytes will be 0xFF. Returns error code. */
WOFS_EXTERN int8_t wofs__alloc_inode(WOFS_STATE_ARG
	wofs_ino_t* const out_ino);

/* Get a temporary block for internal operations. This will not modify */
/* the free block counter, and the block must be erased after use. */
WOFS_EXTERN int8_t wofs__get_temp_block(WOFS_STATE_ARG
	wofs_addr_t* const addr);

/* Find a directory entry in the specified directory. Optionally returns */
/* the address of the first free directory entry slot in which the searched */
/* directory entry would fit. */
WOFS_EXTERN int8_t wofs__find_dirent(WOFS_STATE_ARG wofs_ino_t* const out_ino,
	wofs_ino_t dir, const uint8_t namelen, const char* const name,
	struct wofs_direnthint* const out_hint);

/* Add the specified directory entry (which consists of 'ino', 'namelen' */
/* and 'name') into the directory 'dir'. Hint can be optionally provided */
/* for faster operation (do not handcraft a hint! only hints returned by */
/* wofs__find_dirent() are accepted). The last filled field is 'd_namelen'. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs__add_dirent(WOFS_STATE_ARG wofs_ino_t dir,
	const wofs_ino_t ino, const uint8_t namelen, const char* const name,
	const struct wofs_direnthint* const hint);

/* Remove dead directory entries from directory 'dir'. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs__cleanup_directory(WOFS_STATE_ARG wofs_ino_t dir);

/* Initialize the directory iterator. This function cannot fail, it only */
/* fills out some fields in 'dir'. Hint tracking is enabled if 'namelen' */
/* is in the accepted name length range, and 'out_hint' is not NULL. */
WOFS_EXTERN void wofs__dir_open(WOFS_STATE_ARG struct wofs_dir* const dir,
	wofs_ino_t ino, const uint8_t namelen,
	struct wofs_direnthint* const out_hint);

/* Read the next directory entry. Returns error code, or */
/* '-WOFS_ERR_ENOMORE' if reached the end of directory. */
WOFS_EXTERN int8_t wofs__dir_read(WOFS_STATE_ARG struct wofs_dir* const dir);

/* Check whether the specified inode is a directory and it is empty. This */
/* function is a callback of remove_object. Returns 0 on success, error */
/* code otherwise (including the not-empty error). */
WOFS_EXTERN int8_t wofs__check_dir_RM(WOFS_STATE_ARG wofs_ino_t ino);

/* Check whether the specified inode is valid, and type matches (if 'type' */
/* is not zero). */
WOFS_EXTERN int8_t wofs__check_inode(WOFS_STATE_ARG const wofs_ino_t ino,
	const uint8_t type);

/* This function writes the first part of the inode structure into the */
/* storage leaving it an not "clear" inode. This inode will be purged */
/* during mount if the operation was interrupted. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs__ino_fill_1(WOFS_STATE_ARG
	const struct wofs_inode* const inode, const wofs_addr_t addr);

/* This function writes the second part of the inode structure, completing */
/* the data written by the part-1 function above. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs__ino_fill_2(WOFS_STATE_ARG
	const struct wofs_inode* const inode, const wofs_addr_t addr);

/* Compile a temporary inode structore (WOFS_INOTYPE_TMPDIR or */
/* WOFS_INOTYPE_TMPFILE). Takes care of broken chaining. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs__compile_temp_inode(WOFS_STATE_ARG
	struct wofs_inode* const temp, const struct wofs_inode* const original,
	const uint8_t i_type);

/* Create a new object in the specified directory. Performs check on 'dir'. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs__create_object(WOFS_STATE_ARG wofs_ino_t* const ino,
	wofs_ino_t dir, const char* const name, const uint8_t i_type,
	int8_t (*init_inode_callback)(WOFS_STATE_ARG wofs_addr_t addr,
	wofs_ino_t cur_ino, wofs_ino_t par_ino));

/* Removes an object and its directory entry from 'ino' directory (no type */
/* check will be performed) if 'chk_rm_callback' returns zero. Returns 0 on */
/* success; error code otherwise (including the error code returned by */
/* the callback). */
WOFS_EXTERN int8_t wofs__remove_object(WOFS_STATE_ARG wofs_ino_t ino,
	const uint8_t namelen, const char* const name,
	int8_t (*chk_rm_callback)(WOFS_STATE_ARG wofs_ino_t ino));

/* Copy storage content from the specified address to another. Overlapping */
/* ranges are not supported. Returns error code. */
WOFS_EXTERN int8_t wofs__copy_storage_content(WOFS_STATE_ARG wofs_addr_t dst,
	wofs_addr_t src, wofs_size_t size);

/* Setup file descriptor based on the specified inode. Caller must ensure */
/* the 'ino' is a file inode. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs__fd_setup(WOFS_STATE_ARG struct wofs_fd* const fd,
	const wofs_ino_t ino);

/* Write the fileinfo structure into the specified block. The 'addr' must */
/* hold the base address of the block. Returns error code. */
WOFS_EXTERN int8_t wofs__fileinfo_write(WOFS_STATE_ARG const wofs_addr_t addr,
	const wofs_off_t start_offs, const wofs_ino_t prev_ino);

/* The actual file writing mechanizm. The purpose of the 'write_it' is to */
/* easily implement the fallocate() function. Returns error code. */
WOFS_EXTERN wofs_ssize_t wofs__write_impl(WOFS_STATE_ARG
	struct wofs_fd* const fd, const void* buf, wofs_size_t size,
	int8_t* const out_ec, const uint8_t write_it);

/* Follow the file chain in forward direction. Updates 'fd'. Returns error */
/*code. */
WOFS_EXTERN int8_t wofs__follow_file_chain__forward(WOFS_STATE_ARG
	struct wofs_fd* const fd);

/* Follow the file chain in backward direction. Updates 'fd'. Returns error */
/*code. */
WOFS_EXTERN int8_t wofs__follow_file_chain__backward(WOFS_STATE_ARG
	struct wofs_fd* const fd);

/* Check whether the specified name is a valid directory entry name. The */
/* length of the name is computed, and stored in 'len' (this must not be */
/* NULL). Returns error code. */
WOFS_EXTERN int8_t wofs__check_name(WOFS_STATE_ARG const char* const name,
	uint8_t* const len);

/* Rewrite the file inode specified by 'fd'. Keeps only 'clen' byte of file */
/* content. Returns error code. */
WOFS_EXTERN int8_t wofs__file_rewrite(WOFS_STATE_ARG struct wofs_fd* const fd,
	const uint16_t clen);

/* Restore the previously created temporary file inode back to is origin. */
/* Returns error code. */
WOFS_EXTERN int8_t wofs__file_restore_temp(WOFS_STATE_ARG
	const wofs_ino_t temp_ino);


/* Read the storage. */

#if WRONCEFS_DIRECT_READ
#  include <string.h>
#  define WOFS_READ_EXTERN(_memory_, _addr_, _size_) \
	memcpy((_memory_), \
		(((uint8_t*)WRONCEFS_BASE_ADDRESS)+(_addr_)), \
		(_size_))

#  define WOFS_READ_LOCAL(_type_, _name_, _addr_) \
	const _type_* const _name_ = \
		(const _type_*)(((uint8_t*)WRONCEFS_BASE_ADDRESS)+(_addr_))

#  define WOFS_LOCAL_FIELD(_name_, _field_) \
	_name_->_field_

#  define WOFS_LOCAL_ADDR(_name_) \
	_name_

#  define WOFS_LOCAL_VALUE(_name_) \
	*_name_

#else /* WRONCEFS_DIRECT_READ */
#  define WOFS_READ_EXTERN(_memory_, _addr_, _size_) \
	do \
	{ \
		if(0!=wroncefs_op_read(PASS_STATE (_memory_), (_addr_), (_size_))) \
		{ \
			return -WOFS_ERR_EIOR; \
		} \
	} \
	while(0)

#  define WOFS_READ_LOCAL(_type_, _name_, _addr_) \
	_type_ _name_; \
	do \
	{ \
		if(0!=wroncefs_op_read(PASS_STATE &(_name_), (_addr_), \
			sizeof(_type_))) \
		{ \
			return -WOFS_ERR_EIOR; \
		} \
	} \
	while(0)

#  define WOFS_LOCAL_FIELD(_name_, _field_) \
	_name_._field_

#  define WOFS_LOCAL_ADDR(_name_) \
	&_name_

#  define WOFS_LOCAL_VALUE(_name_) \
	_name_

#endif /* WRONCEFS_DIRECT_READ */

#endif /* WRONCEFS_H */
