/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>
#include <string.h>

static int8_t seek_start(WOFS_STATE_ARG struct wofs_fd* const fd)
{
	WOFS_READ_LOCAL(struct wofs_inode, inode, fd->fd_addr);

	const wofs_ino_t first = WOFS_LOCAL_FIELD(inode, i_first);

	if(WOFS_END_INODE==first)
	{
		return -WOFS_ERR_ENOTRECOVERABLE;
	}

	{
		const int8_t ec = wofs__check_inode(PASS_STATE first,
			WOFS_INOTYPE_FILE);

		if(0!=ec)
		{
			return ec;
		}
	}

	return wofs__fd_setup(PASS_STATE fd, first);
}

static int8_t seek_end(WOFS_STATE_ARG struct wofs_fd* const fd)
{
	for(;;)
	{
		const int8_t ec = wofs__follow_file_chain__forward(PASS_STATE fd);

		if((-WOFS_ERR_ENOMORE)==ec)
		{
			break;
		}
		else
		if(0!=ec)
		{
			return ec;
		}
	}

	uint16_t diff = fd->fd_clen - fd->fd_lpos;

	fd->fd_fpos += diff;
	fd->fd_lpos += diff;

	return 0;
}

WOFS_IMPL wofs_off_t wofs_seek(WOFS_STATE_ARG struct wofs_fd* const fd,
	wofs_off_t offs, uint8_t whence, int8_t* const out_ec)
{
	int8_t ec = 0;

	switch(whence)
	{
	case WOFS_SEEK_END:
		if(0!=(ec=seek_end(PASS_STATE fd)))
		{
			goto error_exit;
		}
		/* fall thru' */
	case WOFS_SEEK_CUR:
		offs = fd->fd_fpos + offs;
		/* fall thru' */
	case WOFS_SEEK_SET:
		/* no op */
		break;

	default:
		ec = -WOFS_ERR_EINVAL;
		goto error_exit;
	}

	if(0>offs)
	{
		ec = -WOFS_ERR_EINVAL;
		goto error_exit;
	}

	/* select start position (0 or fd_fpos) */

	/* |-----------,---------|   */
	/* 0     ^     fpos      end */
	/*       |                   */
	/*       fpos/2, the turning point */

	if((fd->fd_fpos/2)>offs)
	{
		if(0!=(ec=seek_start(PASS_STATE fd)))
		{
			goto error_exit;
		}
	}

	while(fd->fd_fpos<offs)
	{
		wofs_size_t part = fd->fd_clen - fd->fd_lpos;

		if(0==part)
		{
			if(0!=(ec=wofs__follow_file_chain__forward(PASS_STATE fd)))
			{
				if((-WOFS_ERR_ENOMORE)==ec)
				{
					ec = 0;
				}

				break;
			}

			continue;
		}

		wofs_size_t diff = offs - fd->fd_fpos;

		if(part>diff)
		{
			part = diff;
		}

		fd->fd_lpos += part;
		fd->fd_fpos += part;
	}

	while(fd->fd_fpos>offs)
	{
		wofs_size_t part = fd->fd_lpos;

		if(0==part)
		{
			if(0!=(ec=wofs__follow_file_chain__backward(PASS_STATE fd)))
			{
				break;
			}

			continue;
		}

		wofs_size_t diff = fd->fd_fpos - offs;

		if(part>diff)
		{
			part = diff;
		}

		fd->fd_lpos -= part;
		fd->fd_fpos -= part;
	}

	if(NULL!=out_ec)
	{
		*out_ec = ec;
	}

	return fd->fd_fpos;

error_exit:;

	if(NULL!=out_ec)
	{
		*out_ec = ec;
	}

	return -1;
}
