/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "wroncefs.h"
#include "io.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>

uint8_t test_fs_buffer[
#if WRONCEFS_16BIT_MODE
	64*1024
#else /* WRONCEFS_16BIT_MODE */
	128*1024
#endif /* WRONCEFS_16BIT_MODE */
];

static size_t fs_size;
static size_t fs_blksize;
static size_t die_count;
static size_t die_limit;

size_t io_get_size(void)
{
	return fs_size;
}

int io_set_size(size_t size)
{
	if(sizeof(test_fs_buffer)<size)
	{
		return -1;
	}

	fs_size = size;

	return 0;
}

size_t io_get_max_size(void)
{
	return sizeof(test_fs_buffer);
}

size_t io_get_blksize(void)
{
	return fs_blksize;
}

int io_set_blksize(size_t blksize)
{
	int8_t sh = wofs_off_to_shift(blksize);

	if(WOFS_MIN_BLOCKSZSH>sh || WOFS_MAX_BLOCKSZSH<sh)
	{
		return -1;
	}

	fs_blksize = blksize;

	return 0;
}

void io_mass_erase(void)
{
	memset(test_fs_buffer, 0xFF, fs_size);
}

int io_save_data(const char* const filename)
{
	FILE* const file = fopen(filename, "wb");

	if(NULL==file)
	{
		return -1;
	}

	const size_t num = fwrite(test_fs_buffer, fs_size, 1, file);
	const int    err = errno;

	fclose(file);

	if(1!=num)
	{
		errno = err;
		return -1;
	}

	return 0;
}

void io_die_after(size_t bytes)
{
	die_count = 0;
	die_limit = bytes;
}

void io_die_reset(void)
{
	die_limit = 0;
}

int8_t wroncefs_op_write(const void* buf, wofs_addr_t addr, wofs_size_t size)
{
	int result = 0;

	if(fs_size<=(size_t)addr || fs_size<(((size_t)addr)+size))
	{
		return -1;
	}

	if(0<die_limit)
	{
		if(die_count<die_limit)
		{
			size_t rem = die_limit - die_count;

			if(size>rem)
			{
				result = -3;
				size = rem;
			}

			die_count += size;
		}
		else
		if(die_count==die_limit)
		{
			result = -4;
		}
	}

#if !WRONCEFS_CAN_ALTER
	{
		uint8_t* ptr = test_fs_buffer + addr;
		uint8_t* const end = ptr + size;

		for(;ptr<end;++ptr)
		{
			if(0xFF!=(*ptr))
			{
				return -2;
			}
		}
	}
#endif /* !WRONCEFS_CAN_ALTER */

	memcpy(test_fs_buffer+addr, buf, size);

	return result;
}

int8_t wroncefs_op_erase(wofs_addr_t addr, wofs_size_t size)
{
	int result = 0;

	if(fs_size<=(size_t)addr || fs_size<(((size_t)addr)+size))
	{
		return -1;
	}

	if(0<die_limit)
	{
		if(die_count<die_limit)
		{
			size_t rem = die_limit - die_count;

			if(size>rem)
			{
				result = -3;
				size = rem;
			}

			die_count += size;
		}
		else
		if(die_count==die_limit)
		{
			result = -4;
		}
	}

	memset(test_fs_buffer+addr, 0xFF, size);

	return result;
}

#if !WRONCEFS_DIRECT_READ
int8_t wroncefs_op_read(void* buf, wofs_addr_t addr, wofs_size_t size)
{
	if(fs_size<=(size_t)addr || fs_size<(((size_t)addr)+size))
	{
		return -1;
	}

	memcpy(buf, test_fs_buffer+addr, size);

	return 0;
}
#endif /* !WRONCEFS_DIRECT_READ */
